/*************************************************************************************************************
Main.c

This is the main source code file for the Camper Meter

    Version 1.0   Nov 2007

    Copyright 2007 Geoff Graham - http://geoffg.net
    This program is free software: you can redistribute it and/or modify it under the terms of the GNU General
    Public License as published by the Free Software Foundation, either version 2 of the License, or (at your
    option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.  You should have received a copy of the GNU General Public License along with this program.
    If not, see <http://www.gnu.org/licenses/>.


Program summary:
 - Timer 0 is configured to interrupt every 4 mSec (approx).  This is used to multiplex the LEDs in 2 groups of 5.
   e.g.  every 4 mS a new LED in a group is lit.  After all 5 LEDS in the group have had their turn the process
   repeats (20 mS cycle).  The two groups of 5 LEDs make up the 10 LEDs in the bargraph display.
   
 - Every fifth timer 0 interrupt (e.g.  every 20 mS) a number of routines are executed.  Most of the processing is 
   executed at this time.  This includes:-
        - If the comparator output is high then a unit of water consumption is recorded.  Water consumption is 
          therefor counted in units of 20 mS (i.e.  the pump running for one second will record a count of 50).
        - If the water consumption display has changed then the new value is written to EEPROM.
        - Detects if the Voltage button is depressed and if so reads the voltage and scales it for display.
        - Provides a simple animation of the display when it changes.  The display rises quickly from zero then
          slows down to emulate an analog display.
        - Detects if the Reset button is depressed and if so causes the display to flash.  After 3 seconds it will
          reset the water consumption count to zero (i.e.  a full tank).
        - Detects if both the Reset and Voltage buttons are depressed and if so stores the current water
          consumption count into EEPROM as the count representing an empty tank.
        - Handles flashing of the LEDs for reset and empty representation.
        - And FINALLY outputs the appropiate bits to turn on/off the LED display.
 
 - The main() function has very little to do.  It initialises everything and then executes a continuous loop
   calculating the water consumption display from the water consumption count and the count representing an empty
   tank.  Because this calculation involves division of 32 bit integers it can take some time to execute and is 
   therefor not included in the timer 0 interrupt routine.

***************************************************************************************************************/


#define 16F88

#ifdef 16F88
    #include <16F88.h>
    #device adc=8
    #fuses INTRC_IO,NOIESO,NOPUT,NOFCMEN,NOWDT,NOCPD,NOWRT,NOPROTECT,NOMCLR,NOBROWNOUT,NOLVP,NODEBUG
    // Clock Related Fuses (select one):
    //        LP  Low-Power Crystal
    //        XT  Crystal/Resonator
    //        HS  High-Speed Crystal/Resonator
    //        RC  External Resistor-Capacitor (RC) with FOSC/4 output on RA6
    //        RC_IO  External Resistor-Capacitor with I/O on RA6
    //        INTRC - Internal Oscillator with FOSC/4 output on RA6 and I/O on RA7
    //        INTRC_IO - Internal Oscillator with I/O on RA6 and RA7
    //        EC_IO  External Clock with I/O on RA6.
    // Int/Ext Clock Switchover:     IESO,NOIESO
    // Fail-Safe Clock Monitor:      FCMEN,NOFCMEN
    // Watch Dog Timer:              WDT, NOWDT
    // Protect Data Memory:          CPD,NOCPD
    // Protect Code:                 PROTECT,NOPROTECT
    // Prog mem write protected      WRT,NOWRT
    // Use MCLR Pin for Reset:       MCLR, NOMCLR
    // Power-up Timer:               PUT,NOPUT
    // Brown-out Detect:             BROWNOUT,NOBROWNOUT
    // Low-Volt Programming Enable:  LVP, NOLVP
    // In-Circuit Debugger Mode:     DEBUG,NODEBUG
    // CCP1 Pin Selection:           CCPB0, CCPB3

    // processor specific hardware registers
    #byte indf = 0                  // Indirect register
    #byte tmr0 = 1                  // timer count register
    #byte status = 3                // The status register
    #byte fsr = 4                   // Index to RAM
    #byte porta = 5                 // Port A I/O register
    #byte portb = 6                 // Port B I/O register
    #byte pclath = 0x0a             // Program counter high bits
    #byte intcon = 0x0b             // Basic interrupt control
    #byte cmcon = 0x9c              // Analog comparator control

    #use delay(clock=4000000)
#endif

#include "..\Library\StdDefines.h"
#use fast_io(A) 
#use fast_io(B)

#define TRISA               0b00100111
#define TRISB               0b11000000

#define PIN_VOLT_SWITCH     PIN_B7      // input used to detect if volt switch pressed (low = pressed)
#define PIN_RESET_SWITCH    PIN_B6      // input used to detect if reset switch pressed (low = pressed)

#define VOLT_ZERO_OFFSET    146         // For Geoff & Lynda's Camper Meter
//#define VOLT_ZERO_OFFSET  144         // For Jane & Robin's Camper Meter

/************************************************************************************************
Global memory locations
************************************************************************************************/

// RAM storage for the configuration value
#define CONFIG_DATA_START 0x20
#define INT32_DATA_SIZE 4
unsigned int32 tank_empty;
#locate tank_empty = CONFIG_DATA_START

// RAM storage for the water consumption value
// NOTE:  This MUST be contiguous with tank_empty above
unsigned int32 tank_count;
#locate tank_count = CONFIG_DATA_START + INT32_DATA_SIZE

byte delay_count, reset_timer;
byte mask, hi_bank, lo_bank;
byte val, dval;
byte key_volt = OFF;
byte flash_timer;
word flash_leds;

byte water, newwater;

// initilise the eeprom data
// tank_empty = 72,605 or exactly 25 minutes and tank_count = tank_empty
#rom  0x2100 = {0x9d, 0x1b, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00} 

    
/************************************************************************************************
Timer0 interrupt
************************************************************************************************/
#int_timer0
void timer0_interrupt() {
    byte i;
    
    mask >>= 1;
    if(mask == 0) {
        // this section of code executes every 20mS approx (e.g.  at 50Hz)
        mask = 0b00010000;
        delay_count--;
        flash_timer++;

        // if the comparator output is high, record the water consumption
        if((cmcon & 0b10000000) == 0b10000000) tank_count++;
        
        if(water != newwater) {
            newwater = water;
            // write the current count (stored in contiguous RAM) to the eeprom
            for(i = INT32_DATA_SIZE ; i < INT32_DATA_SIZE * 2 ; i++) write_eeprom (i, *(CONFIG_DATA_START + i));
        }   

        if(input(PIN_RESET_SWITCH))
        // if the volt switch is pressed (and not the reset switch) then we just display the adc reading
        //   on the prototype circuit the adc will give VOLT_ZERO_OFFSET for 11.4V which is zero capacity
        //   and VOLT_ZERO_OFFSET + 20 for 12.9V which is full capacity (when battery is idle)
        // NOTE: The voltage is calculated here because when it was put in the while(1) loop in main()
        //       it gave random readings - due to interrupts perhaps ???
            if(!input(PIN_VOLT_SWITCH)) {
                if(!key_volt) {                 // has the volt switch just been pressed?
                    key_volt = ON;              // if yes - record the fact
                    dval = 0;                   //        - and set the display to zero to indicate change
                }
                val = read_adc();               // get the voltage
                if(val < VOLT_ZERO_OFFSET)
                    val = 0;                    // set at zero if it is off the bottom of the range
                else
                    val = (val - VOLT_ZERO_OFFSET) >> 1; // otherwise subtract sero offset and scale (i.e. divide by 2)
                                                // This gives zero for an input of 11.4V and 10 for 12.9V 
                if(val > 10) val = 10;          // finally, limit the maximum to 10
                flash_leds = OFF;
                if(val == 0) flash_leds = 1;
            } else {
                if(key_volt) {                  // has the volt switch been pressed BUT now has been released?
                    key_volt = OFF;             // if yes - record the fact
                    dval = 0;                   //        - and set the display to zero to indicate change
                    }
                val = water;
                flash_leds = OFF;
                if(val == 0) flash_leds = 1;
            }
        else
        // the reset key is down - turn off flashing as later code will turn it on if necessary
                flash_leds = OFF;
        
        // this code is concerned with taking val (the value we wish to display) and copying it to dval (the
        // value on the leds).  To imitate an analog display dval will gradually rise to the level of val.
        
        // if the reading is less than the current display then immediately show the new reading
        if(dval >= val) {
            dval = val;
            delay_count = 1;                    // no animation
            }
        // if the reading is more than the current display then animate by gradually increase the display value
        else
            if(delay_count == 0) {
                dval++;                         // increase the display value
                if(val - dval > 4)
                    delay_count = 1;            // fast animation
                else if(val - dval > 2)
                    delay_count = 2;            // slower animation
                else
                    delay_count = 4;            // slowest animation for the last couple of leds
                }
        
        // the reset switch is held dowm
        // if the switch is held down this code will override the values for dval set in the above code
        // the main function of this code is to flash the leds off/on and reset the water capacity parameters
        if(!input(PIN_RESET_SWITCH)) {
            if(reset_timer < 150) {             // we wait for 3 sec (150 * 20mS)
                flash_leds = 0b1111111111;      // flash all leds
                reset_timer++;
             } else
                if(reset_timer == 150) {        // if the 3 sec has expired (and the reset key is still down)
                    reset_timer++;              // don't execute this code again while the key is down
                    if(input(PIN_VOLT_SWITCH)) {
                        // set the water tank to full - this is done by zeroing the counter
                        water = val = 10;
                        dval = 0;
                        tank_count = 0;         // reset to full tank
                        flash_leds = OFF;
                    } else {
                         // reset the zero IF the pump has been running for more than 60 seconds
                         // this is done by making the empty tank count equal the current count
                         if(tank_count > 3000) tank_empty = tank_count;
                         water = val = dval = 0;
                        flash_leds = 1;
                         // write the configuration variable (stored in contiguous RAM) to the eeprom
                         for(i = 0 ; i < INT32_DATA_SIZE ; i++) write_eeprom (i, *(CONFIG_DATA_START + i));
                    }
                }
        } else {
            if(reset_timer > 0) flash_leds = OFF;
            reset_timer = 0;                    // reset key is up, so set the timer back to zero
        }
        
        // work out what LEDS will be lit up
        // first, if we are flashing the leds, use the value of flash_leds to determine what leds will light up
        if(flash_leds > 0) {
            if((flash_timer & 0b0010000) == 0b0010000) {
                hi_bank = flash_leds >> 5;
                lo_bank = 0b00011111 & flash_leds;
            } else {
                hi_bank = 0;
                lo_bank = 0;
            }
        } else {
            // if we are not flashing the leds we use the value of dval to determine what leds will light up
            if(dval > 5) {
                hi_bank = 0b00011111 >> (10 - dval);
                lo_bank = 0b00011111;
            } else {
                hi_bank = 0;
                lo_bank = 0b00011111 >> (5 - dval); 
            }
        }
    }
    
    // output to the LEDs using the mask var to multiplex the LEDs in 2 banks
    output_b((hi_bank & mask) | (((lo_bank & mask)<<1) & 0b00100000));
    output_a((((lo_bank & mask)<<3) & 0b00011000) | (((lo_bank & mask)<<4) & 0b11000000));
}
    

/**********************************************************************************************
Main program
**********************************************************************************************/

void main() {
    byte i;

    // run this chip at 4MHz using the internal clock
    setup_oscillator(OSC_4MHZ|OSC_INTRC);       
    
    // disable any features of the chip not required in this application
    setup_spi(SPI_SS_DISABLED);
    setup_timer_1(T1_INTERNAL|T1_DIV_BY_1);
    setup_timer_2(T2_DISABLED,0,1);
    setup_ccp1(CCP_OFF);
    setup_vref(FALSE);
    
    // setup timer 0 to interrupt every 4mS (approx)
    setup_timer_0(RTCC_INTERNAL|RTCC_DIV_16);
    enable_interrupts(INT_TIMER0);
    enable_interrupts(GLOBAL);

    // setup PIN_A0 as the analog input
    setup_adc_ports(sAN0|VSS_VDD);
    setup_adc(ADC_CLOCK_INTERNAL);
    set_adc_channel(0);
    
    // setup PIN_A1 and A2 as the comparitor inputs
    setup_comparator(NC_NC_A1_A2);

    // set pin directions and enable pullups on port B
    set_tris_a(TRISA);
    set_tris_b(TRISB);
    port_b_pullups(1);
    
    // get stored value for tank_empty and tank_count
    // this reads 8 bytes (i.e. two int32 variables) from eeprom into RAM
    for(i = 0 ; i < INT32_DATA_SIZE * 2 ; i++) *(CONFIG_DATA_START + i) = read_eeprom(i);
    
    while(forever) {    
        if(tank_count < tank_empty) {
            water = (byte)((((tank_empty + (tank_empty/20) - tank_count) * 10) / tank_empty) & 0xff);
            if(water > 10) water = 10;
        } else
            water = 0;
    }
}
